/* $Id: repo.c 104 2023-05-18 23:43:15Z nishi $ */
#include <swinehub/repo.h>

#include <swinehub.h>
#include <swinehub/base64.h>
#include <swinehub/config.h>
#include <swinehub/util.h>
#include <swinehub/user.h>
#include <swinehub/db.h>
#include <swinehub/language.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>

#define WRITE(fd, data) write(fd, data, strlen(data))

extern char* umethod;
extern char* swinehub_request;
extern char* swinehub_body;

EXTERN_DB;

void swinehub_repo(struct swinehub_config* config){
	char* slash = getenv("PATH_REPLACED") + 2;
	int i;
	char matched = 0;
	int index = -1;
	int index2 = -2;
	for(i = 0; slash[i] != 0; i++){
		if(slash[i] == '/'){
			if(index == -1) index = i;
			slash[i] = 0;
			matched = 1;
			if(index2 == -1){
				index2 = i;
				break;
			}
			index2 = -1;
		}
	}
	char* username = slash;
	char* reponame;
	char* svnpath = strdup("/");
	if(matched){
		username = slash;
		reponame = slash + index + 1;
		if(index2 != -1){
			free(svnpath);
			svnpath = sstrcat("/", slash + index2 + 1);
		}
	}
	if(strcmp(getenv("ISDAV"), "YES") == 0){
		printf("Status: 301 Moved Permanently\r\n");
		printf("Location: %s/%s-%s\r\n", config->prefix, username, reponame);
		printf("\r\n");
		return;
	}
	if(matched){
		if(strcmp(getenv("ISDAV"), "NO") == 0){
			int rev = -1;
			int status;
recheckout:
			{
				char* localpath = sstrcat5(config->path, "/", username, "-", reponame);
				struct stat st;
				if(stat(localpath, &st) != 0){
					free(localpath);
					printf("Repository not found.\n");
					return;
				}
				free(localpath);
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(config->cachepath);
					char* tmp = sstrcat4("file://", config->path, "/", username);
					char* checkoutpath = sstrcat3(tmp, "-", reponame);
				free(tmp);
					execl(config->svn, config->svn, "co", checkoutpath, NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
			}
			{
				char* tmp = sstrcat5(config->cachepath, "/", username, "-", reponame);
				char* chdirpath = strdup(tmp);
				free(tmp);
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(chdirpath);
					execl(config->svn, config->svn, "up", NULL);
					_exit(-1);
				}else{
					waitpid(pid, &status, 0);
				}
				free(chdirpath);
			}
			if(WIFEXITED(status)){
				if(WEXITSTATUS(status) != 0){
					char* tmp = sstrcat5(config->cachepath, "/", username, "-", reponame);
					char* rmpath = strdup(tmp);
					free(tmp);
					int pid = fork();
					if(pid == 0){
						int nullfd = open("/dev/null", O_RDWR);
						dup2(nullfd, fileno(stdout));
						dup2(nullfd, fileno(stderr));
						execl("/bin/rm", "/bin/rm", "-rf", rmpath, NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					free(rmpath);
					goto recheckout;
				}
			}
			char* tmp = sstrcat3(config->path, "/", username);
			char* tmp2 = sstrcat3(tmp, "-", reponame);
			char* str = sstrcat(tmp2, "/db/current");
			free(tmp);
			free(tmp2);
			struct stat s;
			if(stat(str, &s) == 0){
				FILE* f = fopen(str, "r");
				char* buf = malloc(s.st_size + 1);
				buf[s.st_size] = 0;
				fread(buf, 1, s.st_size, f);
				rev = atoi(buf);
				printf("<input type=\"hidden\" value=\"%s\" id=\"repouser\">\n", username);
				printf("<input type=\"hidden\" value=\"%s\" id=\"reponame\">\n", reponame);
				printf("<input type=\"hidden\" value=\"%s\" id=\"svnprefix\">\n", config->prefix);
				printf("<input type=\"hidden\" value=\"%s\" id=\"path\">\n", svnpath);
				char* prefix = sstrcat5(getenv("SCRIPT_NAME"), "/~", username, "/", reponame);
				SW_REPO_TITLE;
				/*printf("<hr>\n");*/
				/* <span id=\"set-readme\" style=\"display: none;\"> | " SW_SET_README ": <input type=\"file\" id=\"readme-file\" onchange=\"swinehub_readme()\"></span> got removed because it's no more functional. */
				/* printf("<h3 class='widthfloat'>README</h3>\n"); */
				char* tmppath = sstrcat5(config->cachepath, "/", username, "-", reponame);
				char* readme = sstrcat(tmppath, "/README." SW_SUFFIX);
				FILE* freadme = fopen(readme, "r");
				if(freadme == NULL){
					free(readme);
					readme = sstrcat(tmppath, "/README");	
					freadme = fopen(readme, "r");
				}
				free(tmppath);
				printf("<div class='box widthfloat'>");
				if(freadme != NULL){
					printf("<pre><code>");
					char ch;
					while(1){
						fread(&ch, 1, 1, freadme);
						if(feof(freadme)) break;
						if(ch == '<'){
							printf("&lt;");
						}else if(ch == '>'){
							printf("&gt;");
						}else{
							printf("%c", ch);
						}
					}
					fclose(freadme);
					printf("</code></pre>");
				}else{
					printf("<center><i class='phantom'>README not found.</i></center>\n");
				}
				printf("</div>");
				printf("<hr>\n");
				if(strcmp(svnpath, "/") == 0){
					char* path = sstrcat7(config->cachepath, "/", username, "-", reponame, "/", svnpath);
					char* commit = malloc(1);
					commit[0] = 0;
					fflush(stdout);
					int* pipes = malloc(sizeof(*pipes) * 2);
					pipe(pipes);
					int pid = fork();
					if(pid == 0){
						close(pipes[0]);
						dup2(pipes[1], fileno(stdout));
						chdir(path);
						execl(config->svn, config->svn, "log", "-r", "HEAD", NULL);
						_exit(-1);
					}else{
						close(pipes[1]);
						char ch;
						commit[0] = 0;
						int counter = 0;
						int newline = 0;
						char truncated;
						while(1){
							int status = read(pipes[0], &ch, 1);
							if(status == 0 && kill(pid, 0) == 0) break;
							char cond = newline < 1;
							if(ch != '\n' && cond == 1){
								continue;
							}else if(ch == '\n' && cond == 1){
								newline++;
								continue;
							}
							counter++;
							if(ch == '<'){
								char* tmp = commit;
								commit = sstrcat(tmp, "&lt;");
								free(tmp);
							}else if(ch == '>'){
								char* tmp = commit;
								commit = sstrcat(tmp, "&gt;");
								free(tmp);
							}else if(ch == '&'){
								char* tmp = commit;
								commit = sstrcat(tmp, "&amp;");
								free(tmp);
							}else{
								char* tmp = commit;
								char* buffer = malloc(2);
								buffer[0] = ch;
								buffer[1] = 0;
								commit = sstrcat(tmp, buffer);
								free(tmp);
								free(buffer);
							}
							if(counter == 1024){
								char* tmp = commit;
								commit = sstrcat(tmp, " ... Truncated!");
								free(tmp);
								truncated = 1;
								break;
							}
						}
						close(pipes[0]);
						free(pipes);
						waitpid(pid, 0, 0);
						if(truncated != 1){
							int j;
							for(j = strlen(commit) - 2; commit[j] != '\n'; j--){
								commit[j] = 0;
							}
						}
					}
					SW_REPO_LATEST_COMMIT;
					free(path);
					free(commit);
				}else{
					char* path = sstrcat5(config->cachepath, "/", username, "-", reponame);
					char* commit = malloc(1);
					commit[0] = 0;
					fflush(stdout);
					int* pipes = malloc(sizeof(*pipes) * 2);
					pipe(pipes);
					int pid = fork();
					if(pid == 0){
						close(pipes[0]);
						dup2(pipes[1], fileno(stdout));
						chdir(path);
						execl(config->svn, config->svn, "log", "-r", "COMMITTED", sstrcat(".", svnpath), NULL);
						_exit(-1);
					}else{
						close(pipes[1]);
						char ch;
						commit[0] = 0;
						int counter = 0;
						int newline = 0;
						char truncated;
						while(1){
							int status = read(pipes[0], &ch, 1);
							if(status == 0 && kill(pid, 0) == 0) break;
							char cond = newline < 1;
							if(ch != '\n' && cond == 1){
								continue;
							}else if(ch == '\n' && cond == 1){
								newline++;
								continue;
							}
							counter++;
							if(ch == '<'){
								char* tmp = commit;
								commit = sstrcat(tmp, "&lt;");
								free(tmp);
							}else if(ch == '>'){
								char* tmp = commit;
								commit = sstrcat(tmp, "&gt;");
								free(tmp);
							}else if(ch == '&'){
								char* tmp = commit;
								commit = sstrcat(tmp, "&amp;");
								free(tmp);
							}else{
								char* tmp = commit;
								char* buffer = malloc(2);
								buffer[0] = ch;
								buffer[1] = 0;
								commit = sstrcat(tmp, buffer);
								free(tmp);
								free(buffer);
							}
							if(counter == 1024){
								char* tmp = commit;
								commit = sstrcat(tmp, " ... Truncated!");
								free(tmp);
								truncated = 1;
								break;
							}
						}
						close(pipes[0]);
						free(pipes);
						waitpid(pid, 0, 0);
						if(truncated != 1){
							int j;
							for(j = strlen(commit) - 2; commit[j] != '\n'; j--){
								commit[j] = 0;
							}
						}
					}
					SW_FILE_LATEST_COMMIT;
					free(path);
				}
				printf("<hr>\n");
				char* _path = sstrcat5(config->cachepath, "/", username, "-", reponame);
				char* path = sstrcat(_path, svnpath);
				free(_path);
				struct stat st;
				if(stat(path, &st) == 0){
					if(S_ISDIR(st.st_mode)){
						printf("<table class='box widthfloat' border=\"0\" id=\"files\">\n");
						printf("<thead>\n");
						printf("	<tr>\n");
						printf("		<th style=\"width: 100%%; \">" SW_NAME "</th>\n");
						printf("		<th style=\"min-width: 150px;\">" SW_TYPE "</th>\n");
						printf("		<th style=\"min-width: 250px;\">" SW_SIZE "</th>\n");
						printf("	</tr>\n");
						printf("</thead>\n");
						printf("<tbody>\n");
						if(strcmp(svnpath, "/") != 0){
							printf("<tr>\n");
							printf("	<td><a href=\"../\">../</a>\n");
							printf("	<td>Directory</td>\n");
							printf("	<td></td>\n");
							printf("</tr>\n");
						}
						struct dirent* d;
						DIR* dir;
						char matchch = '/';
doagain:
					        dir = opendir(path);
						if(dir != NULL){
							while((d = readdir(dir)) != NULL){
								if(strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0 || strcmp(d->d_name, ".svn") == 0) continue;
								char* filepath = sstrcat3(path, "/", d->d_name);
								struct stat s;
								if(stat(filepath, &s) == 0){
									char ch = ' ';
									char* start = sstrcat3("<a href=\"./", d->d_name, "/\">");
									char* end = strdup("</a>");
									char* type = strdup("File");
									if(S_ISDIR(s.st_mode)){
										ch = '/';
										free(type);
										type = strdup("Directory");
									}
									if(strcmp(svnpath, "/") == 0 && strcmp(d->d_name, "README") == 0){
										free(type);
										type = strdup("File &lt;Readme&gt;");
									}
									if(matchch == ch){
										printf("<tr>\n");
										printf("	<td>%s%s%c%s</td>\n", start, d->d_name, ch, end);
										printf("	<td>%s</td>\n", type);
										printf("	<td>%ld</td>\n", s.st_size);
										printf("</tr>\n");
									}
									free(start);
									free(end);
									free(type);
								}
								free(filepath);
							}
							closedir(dir);
						}
						if(matchch == '/'){
							matchch = ' ';
							goto doagain;
						}
						printf("</tbody>\n");
						printf("</table>\n");
					}else if(S_ISREG(st.st_mode)){
						int status;
						int pipes[2];
						pipe(pipes);
						int pid = fork();
						if(pid == 0){
							close(pipes[0]);
							int nullfd = open("/dev/null", O_RDWR);
							dup2(pipes[1], fileno(stdout));
							dup2(nullfd, fileno(stderr));
							char* chdirpath = sstrcat5(config->cachepath, "/", username, "-", reponame);
							chdir(chdirpath);
							char* getpath = sstrcat(".", svnpath);
							execl(config->svn, config->svn, "propget", "svn:mime-type", getpath, NULL);
							free(getpath);
							free(chdirpath);
							_exit(-1);
						}else{
							close(pipes[1]);
							waitpid(pid, &status, 0);
							if(WIFEXITED(status)){
								int exitstat = WEXITSTATUS(status);
								if(exitstat != 0){
									printf("<div class='widthfloat box'><pre><code>");
									FILE* f = fopen(path, "r");
									char ch;
									while(1){
										fread(&ch, 1, 1, f);
										if(feof(f)) break;
										if(ch == '<'){
											printf("&lt;");
										}else if(ch == '>'){
											printf("&gt;");
										}else{
											printf("%c", ch);
										}
									}
									fclose(f);
									printf("</code></pre></div>\n");
								}else{
									char* mime = malloc(1);
									mime[0] = 0;
									char* buf = malloc(2);
									buf[1] = 0;
									while(1){
										if(read(pipes[0], buf, 1) == 0) break;
										if(buf[0] == '\n') continue;
										char* tmpbuf = mime;
										mime = sstrcat(tmpbuf, buf);
										free(tmpbuf);
									}
									free(buf);
									printf("File cannot be shown. (MIME: <code>%s</code>)\n", mime);
									free(mime);
								}
							}
						}
					}
				}else{
					printf("<tr>\n");
					printf("	<td>File/Directory not found</td>\n");
					printf("	<td></td>\n");
					printf("</tr>\n");
				}
				free(path);
				fclose(f);
				free(buf);
			}else{
				printf("Repository not found.");
			}
			free(str);
		}else{
			char* svn_path = sstrcat5(config->path, "/", username, "/", reponame);
			struct stat st;
			if(stat(svn_path, &st) == 0){
				char goahead = 0;
				if(
						(strcmp(umethod, "GET") == 0)
					||	(strcmp(umethod, "PROPFIND") == 0)
					||	(strcmp(umethod, "OPTIONS") == 0)
					||	(strcmp(umethod, "REPORT") == 0)
				){
					goahead = 1;
				}else{
					if(strcmp(getenv("ISAUTH"), "NO") == 0){
						goahead = 0;
					}else{
						goahead = 1;
					}
				}
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
			}
		}
	}else{
		swinehub_user(config);
	}
	free(svnpath);
}

char swinehub_check_repo(struct swinehub_config* config){
	char result = 0;
	char* path = strdup(getenv("PATH_REPLACED")) + 2;
	int i;
	char count = 0;
	for(i = 0; path[i] != 0; i++){
		if(path[i] == '/'){
			path[i] = '-';
			count++;
		}
		if(count == 2){
			path[i] = 0;
			break;
		}
	}
	if(count == 0){
		datum key;
		key.dptr = path;
		key.dsize = strlen(path) + 1;
		datum value = dbm_fetch(swinehub_user_db, key);
		if(value.dptr != NULL) result = 1;
	}else{
		char* svn_path = sstrcat3(config->path, "/", path);
		struct stat s;
		if(stat(svn_path, &s) == 0) result = 1;
	}
	free(path - 2);
	return result;
}
